// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint.metadata;

import 'constraint.dart';
import 'src/preconditions.dart';
import 'src/runtime_constraint_resolver.dart';
import 'package:collection/collection.dart';
import 'package:option/option.dart';
import 'package:quiver/collection.dart';
import 'src/util.dart';

abstract class TypeDescriptorResolver {
  factory TypeDescriptorResolver() = RuntimeTypeDescriptorResolver;

  Option<TypeDescriptor> resolveFor(Type clazz);
}

// Not sure if we need this or not. Java's one has a fair load of stuff but
// we may be able to stick at least some of them on Constraint itself as Dart's
// annotations are much less sucky
class ConstraintDescriptor<T extends Constraint> {
  final T constraint;

  ConstraintDescriptor(this.constraint) {
    ensure(constraint, isNotNull);
  }

  int get hashCode => constraint.hashCode;

  bool operator==(other) => other is ConstraintDescriptor
      && constraint == other.constraint;

  String toString() => constraint.toString();
}

abstract class ElementDescriptor {
  /// includes all constraints inherited from parent classes,
  /// interfaces and mixins
  final Set<ConstraintDescriptor> constraintDescriptors;

  ElementDescriptor(this.constraintDescriptors) {
    ensure(constraintDescriptors, isNotNull);
  }
}

class MemberDescriptor extends ElementDescriptor {
  final Symbol name;
  final Option<TypeDescriptor> typeDescriptor;
  final bool isMultivalued;
  final bool isGeneric;

  MemberDescriptor(this.name, Set<ConstraintDescriptor> constraintDescriptors,
    this.typeDescriptor, {this.isMultivalued: false, this.isGeneric: false})
      : super(constraintDescriptors) {
    ensure(name, isNotNull);
  }

  MemberDescriptor merge(MemberDescriptor other) {
    // TODO: which TypeDescriptor do we use. Is it possible to override a
    // property with a subtype?
    // As long as we always merge so that the most refined type is preserved
    final newConstraints = new Set<ConstraintDescriptor>()
        ..addAll(other.constraintDescriptors)
        ..addAll(constraintDescriptors);
    return new MemberDescriptor(name, newConstraints, typeDescriptor);
  }

  String toString() => 'MemberDescriptor[$name]';

}

/**
 * Provides access to all the constraints on a [type].
 *
 * [Constraint]s on the type itself can be accessed via [constraintDescriptors].
 * These include all constraints inherited from all parent classes, interfaces
 * and mixins.
 *
 * Constraints on members can be accessed via [memberDescriptorsMap] and are
 * similarly inherited.
 */
class TypeDescriptor extends ElementDescriptor {
  final Type type;


  /// includes all member constraints inherited from parent classes,
  /// interfaces and mixins
  final Map<Symbol, MemberDescriptor> memberDescriptorsMap;
  Iterable<MemberDescriptor> get memberDescriptors => memberDescriptorsMap.values;

  TypeDescriptor._internal(this.type,
      Set<ConstraintDescriptor> constraintDescriptors,
        this.memberDescriptorsMap)
      : super(constraintDescriptors) {
    ensure(type, isNotNull);
    ensure(memberDescriptorsMap, isNotNull);
  }

  factory TypeDescriptor(Type type,
      Set<ConstraintDescriptor> constraintDescriptors,
      Set<MemberDescriptor> memberDescriptors,
      Option<TypeDescriptor> superclassDescriptor,
      Iterable<TypeDescriptor> superinterfaceDescriptors,
      { bool isCollection: false, bool isGeneric: false }) {

    ensure(constraintDescriptors, isNotNull);
    ensure(memberDescriptors, isNotNull);
    ensure(superclassDescriptor, isNotNull);

    final Iterable<TypeDescriptor> inheritedDescriptors =
        []..addAll(optionAsIterable(superclassDescriptor))
        ..addAll(superinterfaceDescriptors);

    final mergedMemberDescriptors = _createMergedMemberDescriptorMap(
            inheritedDescriptors, memberDescriptors);

    final mergedConstraintDescriptors = _createMergedConstraintDescriptors(
            inheritedDescriptors, constraintDescriptors);


    return new TypeDescriptor._internal(type, mergedConstraintDescriptors,
        mergedMemberDescriptors);
  }

  static Set<ConstraintDescriptor> _createMergedConstraintDescriptors(
      Iterable<TypeDescriptor> inheritedDescriptors,
      Set<ConstraintDescriptor> constraintDescriptors) {

    final mergedDescs =  new Set<ConstraintDescriptor>.identity();
    final inheritedConstraints = inheritedDescriptors
        .map((td) => td.constraintDescriptors)
        .expand((e) => e);

    mergedDescs
        ..addAll(inheritedConstraints)
        ..addAll(constraintDescriptors);
    return mergedDescs;
  }

  static Map<Symbol, MemberDescriptor> _createMergedMemberDescriptorMap(
      Iterable<TypeDescriptor> inheritedDescriptors,
      Set<MemberDescriptor> memberDescriptors) {

    final mm = new Multimap<Symbol, MemberDescriptor>();

    inheritedDescriptors.forEach((td) =>
        td.memberDescriptorsMap.forEach((k, v) {
      mm.add(k, v);
    }));

    memberDescriptors.forEach((m) {
      mm.add(m.name, m);
    });

    final combinedMap = <Symbol, MemberDescriptor>{};

    mm.toMap().forEach((sym, itofmem) {
      combinedMap[sym] = itofmem.skip(1)
          .fold(itofmem.first, (r, v) => r.merge(v));
    });
    final combinedMemberDescriptors = new UnmodifiableMapView(combinedMap);
    return combinedMemberDescriptors;
  }

  String toString() => 'TypeDescriptor[$type]';

}