// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint.metadata.resolver.runtime;

import '../constraint.dart';
import 'package:constrain/metadata.dart';
import 'dart:mirrors';
import 'package:option/option.dart';
import 'preconditions.dart';
import 'util.dart';

final TypeMirror _iterableType = reflectType(Iterable);
final TypeMirror _mapType = reflectType(Map);

class RuntimeTypeDescriptorResolver implements TypeDescriptorResolver {
  final Map _cache = <Type, Option<TypeDescriptor>>{};

  // TODO: Does Dart have weak references?

  @override
  Option<TypeDescriptor> resolveFor(Type clazz) {
    if (clazz == null) {
      return const None();
    }

    return _resolveForClass(reflectClass(clazz));
  }

  bool _isCollection(TypeMirror tm) =>
      tm is ClassMirror && tm.isSubtypeOf(_iterableType) || tm.isSubtypeOf(_mapType);

  bool _isGeneric(TypeMirror tm) =>  tm is ClassMirror && tm.typeArguments.isNotEmpty;

  Option<TypeDescriptor> _resolveForClass(ClassMirror cm) {
    if (cm == null || !cm.hasReflectedType) {
      return const None();
    }

    final cachedDescriptor = _cache[cm.reflectedType];

    if (cachedDescriptor != null) {
      return cachedDescriptor;
    }

    // TODO: is there any risk of concurrent modification. i.e. different futures
    // triggering this?

    /*
     * To cope with type loops we put a proxy in to start with. If we hit a
     * type loop then we will end up actually linking up the proxy to a member.
     * So when we have the real type, we first wire the real descriptor into
     * the proxy before we replace it in the cache. The vast majority of
     * members will be wired into a real descriptor rather than a proxy
     */
    final proxy = new _TypeDescriptorProxy();
    _cache[cm.reflectedType] = new Some(proxy);

    final newDescriptor = _resolveForClassUncached(cm);
    if (newDescriptor.nonEmpty()) {
      proxy.realTypeDescriptor = newDescriptor.get();
    }

    _cache[cm.reflectedType] = newDescriptor;

    return newDescriptor;
  }

  Option<TypeDescriptor> _resolveForClassUncached(ClassMirror cm) {

    if (cm.isTopLevel && cm.owner.simpleName == #dart.core) {
      return const None();
    }

    final Set<MemberDescriptor> memberDescriptors = _resolveMembers(cm);

    Option<TypeDescriptor> superclassDescriptor = _resolveSuperclass(cm);

    Iterable<TypeDescriptor> superinterfaceDescs = _resolveSuperinterfaces(cm);

    final Set<ConstraintDescriptor> classConstraints =
        _resolveConstraints(cm);

    final hasConstraints =
        memberDescriptors.isNotEmpty
        || classConstraints.isNotEmpty
        || superclassDescriptor.nonEmpty()
        || superinterfaceDescs.isNotEmpty;

    return hasConstraints ?
        new Some(new TypeDescriptor(cm.reflectedType, classConstraints,
            memberDescriptors, superclassDescriptor, superinterfaceDescs))
        : const None();
  }

  Option<TypeDescriptor> _resolveSuperclass(ClassMirror cm) {
    return _resolveForClass(cm.superclass);
  }

  Iterable<TypeDescriptor> _resolveSuperinterfaces(ClassMirror cm) {
    return cm.superinterfaces.map((i) =>
        optionAsIterable(_resolveForClass(i)))
        .expand((e) => e);
  }

  Set<MemberDescriptor> _resolveMembers(ClassMirror cm) {
    final memberDescriptors = new Set<MemberDescriptor>();

    cm.declarations.forEach((k, v) {
      if ((v is VariableMirror && !v.isStatic) ||
          (v is MethodMirror && !v.isStatic && (v.isGetter /* || v.isSetter */))) {
        final Set<ConstraintDescriptor> constraintDescriptors =
            _resolveConstraints(v);
        final TypeMirror rawMemberType = v is VariableMirror ? v.type :
          v is MethodMirror ? v.returnType : throw new StateError('oops');


        final isCollection = _isCollection(rawMemberType);
        final isGeneric = _isGeneric(rawMemberType);

        if (isCollection && !isGeneric) {
          return const None(); // nothing to go on here
        }

        // TODO: not sure what to do about more than one type arg.
        // Maybe this should be treated same as no generics?
        final memberType = !isCollection ? rawMemberType : rawMemberType.typeArguments.first;

        final Option<TypeDescriptor> memberTypeDescriptor =
            _resolveForType(memberType);

        if (constraintDescriptors.isNotEmpty || memberTypeDescriptor.nonEmpty()) {
          memberDescriptors.add(new MemberDescriptor(k, constraintDescriptors,
              memberTypeDescriptor,
              isMultivalued: isCollection, isGeneric: isGeneric));
        }
      }
    });

    return memberDescriptors;
  }

  Option<TypeDescriptor> _resolveForType(TypeMirror tm) {
    return tm is ClassMirror ? _resolveForClass(tm) : const None();
  }

  Set<ConstraintDescriptor> _resolveConstraints(DeclarationMirror dm) {
    return dm.metadata
        .where((InstanceMirror im) => im.reflectee is Constraint)
        .map((im) => im.reflectee)
        .map((c) => new ConstraintDescriptor(c))
        .toSet();
  }
}

class _TypeDescriptorProxy implements TypeDescriptor {
  TypeDescriptor _realTypeDescriptor;

  TypeDescriptor get realTypeDescriptor {
    // to prevent it being called before it's initialised
    ensure(_realTypeDescriptor, isNotNull);
    return _realTypeDescriptor;
  }

  set realTypeDescriptor(TypeDescriptor d) {
    _realTypeDescriptor = d;
  }

  @override
  Set<ConstraintDescriptor> get constraintDescriptors =>
      realTypeDescriptor.constraintDescriptors;

  @override
  Iterable<MemberDescriptor> get memberDescriptors =>
      realTypeDescriptor.memberDescriptors;

  @override
  Map<Symbol, MemberDescriptor> get memberDescriptorsMap =>
    realTypeDescriptor.memberDescriptorsMap;

  @override
  Type get type => realTypeDescriptor.type;

}

// TODO: need to:
// - support inheritence of types
// - difference between simple types and classes
//   - i.e. classes have super types, simple types won't have their own type constraints